@@ -27,9 +27,10 @@ import pyghmi.ipmi.events as sel
 import pyghmi.ipmi.fru as fru
 import pyghmi.ipmi.oem.generic as genericoem
 from pyghmi.ipmi.oem.lookup import get_oem_handler
-import pyghmi.ipmi.private.util as pygutil
+import pyghmi.ipmi.private.util as util
 from pyghmi.ipmi import sdr
 
+
 try:
     from pyghmi.ipmi.private import session
 except ImportError:
@@ -148,7 +149,7 @@ class Command(object):
 
     def __init__(self, bmc=None, userid=None, password=None, port=623,
                  onlogon=None, kg=None, privlevel=None, verifycallback=None,
-                 keepalive=True):
+                 keepalive=True, **kwargs):
         # TODO(jbjohnso): accept tuples and lists of each parameter for mass
         # operations without pushing the async complexities up the stack
         self.onlogon = onlogon
@@ -160,6 +161,7 @@ class Command(object):
         self._netchannel = None
         self._ipv6support = None
         self.certverify = verifycallback
+        self.kwargs = kwargs
         if bmc is None:
             self.ipmi_session = localsession.Session()
         elif onlogon is not None:
@@ -239,6 +241,10 @@ class Command(object):
                 '<I', struct.pack('3B', *response['data'][6:9]) + b'\x00')[0],
             'product_id': struct.unpack(
                 '<H', struct.pack('2B', *response['data'][9:11]))[0],
+            'firmware_version': '{0}.{1}{2}'.format(
+                response['data'][2] & 0b1111111,
+                response['data'][3] >> 4 & 0b1111,
+                response['data'][3] & 0b1111)
         }
 
     def oem_init(self):
@@ -312,7 +318,7 @@ class Command(object):
         self.oem_init()
         self._oem.reseat_bay(bay)
 
-    def set_power(self, powerstate, wait=False):
+    def set_power(self, powerstate, wait=False, bridge_request=None):
         """Request power state change (helper)
 
         :param powerstate:
@@ -328,23 +334,33 @@ class Command(object):
                      requested state change for 300 seconds.
                      If a non-zero number, adjust the wait time to the
                      requested number of seconds
+        :param bridge_request: The target slave address and channel number for
+                               the bridge request.
         :raises: IpmiException on an error
         :returns: dict -- A dict describing the response retrieved
         """
+        self.oem_init()
+
+        if hasattr(self._oem, 'set_power'):
+            return self._oem.set_power(powerstate,
+                                       bridge_request=bridge_request)
+
+        if hasattr(self._oem, 'process_power_state'):
+            powerstate = self._oem.process_power_state(
+                powerstate, bridge_request=bridge_request)
+
         if powerstate not in power_states:
             raise exc.InvalidParameterValue(
                 "Unknown power state %s requested" % powerstate)
         newpowerstate = powerstate
-        response = self.raw_command(netfn=0, command=1)
-        if 'error' in response:
-            raise exc.IpmiException(response['error'])
-        oldpowerstate = 'on' if (response['data'][0] & 1) else 'off'
+        oldpowerstate = self._get_power_state(bridge_request=bridge_request)
         if oldpowerstate == newpowerstate:
             return {'powerstate': oldpowerstate}
         if newpowerstate == 'boot':
             newpowerstate = 'on' if oldpowerstate == 'off' else 'reset'
         response = self.raw_command(
-            netfn=0, command=2, data=[power_states[newpowerstate]])
+            netfn=0, command=2, data=[power_states[newpowerstate]],
+            bridge_request=bridge_request)
         if 'error' in response:
             raise exc.IpmiException(response['error'])
         lastresponse = {'pendingpowerstate': newpowerstate}
@@ -358,10 +374,9 @@ class Command(object):
                 waitpowerstate = newpowerstate
             currpowerstate = None
             while currpowerstate != waitpowerstate and waitattempts > 0:
-                response = self.raw_command(netfn=0, command=1, delay_xmit=1)
-                if 'error' in response:
-                    raise exc.IpmiException(response['error'])
-                currpowerstate = 'on' if (response['data'][0] & 1) else 'off'
+                currpowerstate = self._get_power_state(
+                    delay_xmit=1,
+                    bridge_request=bridge_request)
                 waitattempts -= 1
             if currpowerstate != waitpowerstate:
                 raise exc.IpmiException(
@@ -370,6 +385,16 @@ class Command(object):
         else:
             return lastresponse
 
+    def _get_power_state(self, delay_xmit=None, bridge_request=None):
+        response = self.raw_command(netfn=0, command=1, delay_xmit=delay_xmit,
+                                    bridge_request=bridge_request)
+        if 'error' in response:
+            raise exc.IpmiException(response['error'])
+        assert (response['command'] == 1 and response['netfn'] == 1)
+        curr_power_state = 'on' if (response['data'][0] & 1) else 'off'
+
+        return curr_power_state
+
     def get_video_launchdata(self):
         """Get data required to launch a remote video session to target.
 
@@ -382,9 +407,8 @@ class Command(object):
 
     def reset_bmc(self):
         """Do a cold reset in BMC"""
-
-        response = self.raw_command(netfn=6, command=2)
-        if 'error' in response:
+        response = self.raw_command(netfn=6, command=2, retry=False)
+        if response and 'error' in response:
             raise exc.IpmiException(response['error'])
 
     def set_bootdev(self,
@@ -509,22 +533,25 @@ class Command(object):
                                             rslun=rslun)
         return rsp
 
-    def get_power(self):
+    def get_power(self, bridge_request=None):
         """Get current power state of the managed system
 
         The response, if successful, should contain 'powerstate' key and
         either 'on' or 'off' to indicate current state.
 
+        :param bridge_request: The target slave address and channel number for
+                               the bridge request.
         :returns: dict -- {'powerstate': value}
         """
-        response = self.raw_command(netfn=0, command=1)
-        if 'error' in response:
-            raise exc.IpmiException(response['error'])
-        assert (response['command'] == 1 and response['netfn'] == 1)
-        powerstate = 'on' if (response['data'][0] & 1) else 'off'
-        return {'powerstate': powerstate}
+        self.oem_init()
+        if hasattr(self._oem, 'get_power'):
+            return {'powerstate': self._oem.get_power(
+                bridge_request=bridge_request)}
 
-    def set_identify(self, on=True, duration=None):
+        return {'powerstate': self._get_power_state(
+            bridge_request=bridge_request)}
+
+    def set_identify(self, on=True, duration=None, blink=False):
         """Request identify light
 
         Request the identify light to turn off, on for a duration,
@@ -536,10 +563,14 @@ class Command(object):
         """
         self.oem_init()
         try:
-            self._oem.set_identify(on, duration)
+            self._oem.set_identify(on, duration, blink)
+            return
+        except exc.BypassGenericBehavior:
             return
         except exc.UnsupportedFunctionality:
             pass
+        if blink:
+            raise exc.IpmiException('Blink not supported with generic IPMI')
         if duration is not None:
             duration = int(duration)
             if duration > 255:
@@ -575,10 +606,18 @@ class Command(object):
         # to store cached copies for given device id, product id, mfg id,
         # sdr timestamp, our data version revision, aux firmware revision,
         # and oem defined field
+        self.oem_init()
         if self._sdr is None:
-            self._sdr = sdr.SDR(self, self._sdrcachedir)
+            if hasattr(self._oem, 'init_sdr'):
+                self._sdr = self._oem.init_sdr()
+            else:
+                self._sdr = sdr.SDR(self, self._sdrcachedir)
         return self._sdr
 
+    def get_event_constants(self):
+        self.oem_init()
+        return self._oem.get_oem_event_const()
+
     def get_event_log(self, clear=False):
         """Retrieve the log of events, optionally clearing
 
@@ -615,6 +654,14 @@ class Command(object):
         return sel.EventHandler(self.init_sdr(), self).decode_pet(specifictrap,
                                                                   petdata)
 
+    def get_ikvm_methods(self):
+        self.oem_init()
+        return self._oem.get_ikvm_methods()
+
+    def get_ikvm_launchdata(self):
+        self.oem_init()
+        return self._oem.get_ikvm_launchdata()
+
     def get_inventory_descriptions(self):
         """Retrieve list of things that could be inventoried
 
@@ -648,18 +695,6 @@ class Command(object):
         return self._oem.get_inventory_of_component(component)
 
     def _get_zero_fru(self):
-        # It is expected that a manufacturer matches SMBIOS to IPMI
-        # get system uuid return data.  If a manufacturer does not
-        # do so, they should handle either deletion or fixup in the
-        # OEM processing pass.  Code optimistically assumes that if
-        # data is returned, than the vendor is properly using it.
-        zerofru = fru.FRU(ipmicmd=self).info
-        if zerofru is None:
-            zerofru = {}
-        guiddata = self.raw_command(netfn=6, command=0x37)
-        if 'error' not in guiddata:
-            zerofru['UUID'] = pygutil.decode_wireformat_uuid(
-                guiddata['data'])
         # Add some fields returned by get device ID command to FRU 0
         # Also rename them to something more in line with FRU 0 field naming
         # standards
@@ -668,8 +703,21 @@ class Command(object):
         device_id['Device Revision'] = device_id.pop('device_revision')
         device_id['Manufacturer ID'] = device_id.pop('manufacturer_id')
         device_id['Product ID'] = device_id.pop('product_id')
+
+        zerofru = fru.FRU(ipmicmd=self).info
+        if zerofru is None:
+            zerofru = {}
         zerofru.update(device_id)
-        return self._oem.process_fru(zerofru)
+        zerofru = self._oem.process_zero_fru(zerofru)
+        # If uuid is not returned in OEM processing,
+        # then it is expected that a manufacturer matches SMBIOS to IPMI
+        # get system uuid return data.
+        if 'UUID' not in zerofru:
+            guiddata = self.raw_command(netfn=6, command=0x37)
+            if 'error' not in guiddata:
+                zerofru['UUID'] = util.\
+                    decode_wireformat_uuid(guiddata['data'])
+        return zerofru
 
     def get_inventory(self):
         """Retrieve inventory of system
@@ -690,7 +738,9 @@ class Command(object):
             if fruinf is not None:
                 fruinf = self._oem.process_fru(fruinf,
                                                self._sdr.fru[fruid].fru_name)
-            yield (self._sdr.fru[fruid].fru_name, fruinf)
+            # check the fruinf again as the oem process may return None
+            if fruinf:
+                yield (self._sdr.fru[fruid].fru_name, fruinf)
         for componentpair in self._oem.get_oem_inventory():
             yield componentpair
 
@@ -740,6 +790,18 @@ class Command(object):
             summary['badreadings'] = fallbackreadings
         return summary
 
+    def get_system_power_watts(self):
+        self.oem_init()
+        return self._oem.get_system_power_watts(self)
+    
+    def get_inlet_temperature(self):
+        self.oem_init()
+        return self._oem.get_inlet_temperature(self)
+
+    def get_average_processor_temperature(self):
+        self.oem_init()
+        return self._oem.get_average_processor_temperature(self)
+
     def get_sensor_reading(self, sensorname):
         """Get a sensor reading by name
 
@@ -759,7 +821,7 @@ class Command(object):
                 if 'error' in rsp:
                     raise exc.IpmiException(rsp['error'], rsp['code'])
                 return self._sdr.sensors[sensor].decode_sensor_reading(
-                    rsp['data'])
+                    self, rsp['data'])
         self.oem_init()
         return self._oem.get_sensor_reading(sensorname)
 
@@ -771,7 +833,12 @@ class Command(object):
         one byte, return the int value
         """
         fetchcmd = bytearray((channel, param, 0, 0))
-        fetched = self.xraw_command(0xc, 2, data=fetchcmd)
+        try:
+            fetched = self.xraw_command(0xc, 2, data=fetchcmd)
+        except exc.IpmiException as ie:
+            if ie.ipmicode == 0x80:
+                return None
+            raise
         fetchdata = fetched['data']
         if bytearray(fetchdata)[0] != 17:
             return None
@@ -794,9 +861,9 @@ class Command(object):
         else:
             raise Exception("Unrecognized data format " + repr(fetchdata))
 
-    def get_extended_bmc_configuration(self):
+    def get_extended_bmc_configuration(self, hideadvanced=True):
         self.oem_init()
-        return self._oem.get_extended_bmc_configuration()
+        return self._oem.get_extended_bmc_configuration(hideadvanced=hideadvanced)
 
     def get_bmc_configuration(self):
         self.oem_init()
@@ -826,6 +893,48 @@ class Command(object):
         self.oem_init()
         self._oem.clear_system_configuration()
 
+    def set_net6_configuration(self, static_addresses=None, static_gateway=None, channel=None):
+        if static_addresses is None and static_gateway is None:
+            return
+        if channel is None:
+            channel = self.get_network_channel()
+        if static_addresses is not None:
+            i = 0
+            for va in static_addresses:
+                if '/' in va:
+                    va, plen = va.split('/', 1)
+                else:
+                    plen = '64'
+                plen = int(plen)
+                vab = bytearray(socket.inet_pton(socket.AF_INET6, va))
+                cmddata = bytearray([channel, 56, 0, 0x80]) + vab + bytearray([plen, 0])
+                self.xraw_command(netfn=0xc, command=1, data=cmddata)
+        if static_gateway is not None:
+            gwb = bytearray(socket.inet_pton(socket.AF_INET6, static_gateway))
+            cmddata = bytearray([channel, 65]) + gwb
+            self.xraw_command(netfn=0xc, command=1, data=cmddata)
+    
+    def get_net6_configuration(self, channel=None):
+        if channel is None:
+            channel = self.get_network_channel()
+        retdata = {}
+        ip6a = self.xraw_command(netfn=0xc, command=2, data=(channel, 56, 0, 0))
+        ip6d = bytearray(ip6a['data'])
+        if ip6d[0] != 0x11:
+            raise Exception('Unsupported reply')
+        if ip6d[2] & 0x80 == 0x80:
+            ip6b = ip6d[3:19]
+            ip6addr = socket.inet_ntop(socket.AF_INET6, ip6b)
+            plen = ip6d[19]
+            retdata['static_addrs'] = ['{}/{}'.format(ip6addr, plen)]
+        ip6g = self.xraw_command(netfn=0xc, command=2, data=(channel, 65, 0, 0))
+        ip6gd = bytearray(ip6g['data'])
+        if ip6gd[0] != 0x11:
+            raise Exception('Unsupported reply')
+        gwa = socket.inet_ntop(socket.AF_INET6, bytes(ip6gd[1:17]))
+        retdata['static_gateway'] = gwa
+        return retdata
+
     def set_net_configuration(self, ipv4_address=None, ipv4_configuration=None,
                               ipv4_gateway=None, channel=None):
         """Set network configuration data.
@@ -840,6 +949,9 @@ class Command(object):
         :param ipv4_gateway: IP address of gateway to use.
         :param channel:  LAN channel to configure, defaults to autodetect
         """
+        if (ipv4_address is None and ipv4_configuration is None
+                and ipv4_gateway is None):
+            return
         if channel is None:
             channel = self.get_network_channel()
         if ipv4_configuration is not None:
@@ -928,6 +1040,12 @@ class Command(object):
         :param gateway_macs: Whether to retrieve mac addresses for gateways
         :returns: A dictionary of network configuration data
         """
+        self.oem_init()
+
+        # support for huawei 2288H server
+        if hasattr(self._oem, 'get_oem_net_configuration'):
+            return self._oem.get_oem_net_configuration()
+
         if channel is None:
             channel = self.get_network_channel()
         retdata = {}
@@ -954,7 +1072,7 @@ class Command(object):
             retdata['ipv4_backup_gateway_mac'] = self._fetch_lancfg_param(
                 channel, 15)
         self.oem_init()
-        self._oem.add_extra_net_configuration(retdata)
+        self._oem.add_extra_net_configuration(retdata, channel)
         return retdata
 
     def get_sensor_data(self):
@@ -975,7 +1093,8 @@ class Command(object):
                 if rsp['code'] == 203:  # Sensor does not exist, optional dev
                     continue
                 raise exc.IpmiException(rsp['error'], code=rsp['code'])
-            yield self._sdr.sensors[sensor].decode_sensor_reading(rsp['data'])
+            yield self._sdr.sensors[sensor].\
+                decode_sensor_reading(self, rsp['data'])
         self.oem_init()
         for reading in self._oem.get_sensor_data():
             yield reading
@@ -1049,6 +1168,9 @@ class Command(object):
             channel = self.get_network_channel()
         rqdata = (channel, 0x11, 0, 0)
         rsp = self.xraw_command(netfn=0xc, command=2, data=rqdata)
+        self.oem_init()
+        if hasattr(self._oem, 'get_alert_destination_count'):
+            return self._oem.get_alert_destination_count(ord(rsp['data'][1]))
         return bytearray(rsp['data'])[1]
 
     def get_alert_destination(self, destination=0, channel=None):
@@ -1068,6 +1190,10 @@ class Command(object):
         :param destination:  The destination number.  Defaults to 0
         :param channel: The channel for alerting.  Defaults to current channel
         """
+        self.oem_init()
+        if hasattr(self._oem, 'get_alert_destination'):
+            return self._oem.get_alert_destination(destination, channel)
+
         destinfo = {}
         if channel is None:
             channel = self.get_network_channel()
@@ -1200,27 +1326,13 @@ class Command(object):
         :param channel: The channel to configure the alert on.  Defaults to
                 current
         """
+        self.oem_init()
+        if hasattr(self._oem, 'set_alert_destination'):
+            self._oem.set_alert_destination(ip)
+            return
         if channel is None:
             channel = self.get_network_channel()
-        if ip is not None:
-            destdata = bytearray((channel, 19, destination))
-            try:
-                parsedip = socket.inet_aton(ip)
-                destdata.extend((0, 0))
-                destdata.extend(parsedip)
-                destdata.extend(b'\x00\x00\x00\x00\x00\x00')
-            except socket.error:
-                if self._supports_standard_ipv6:
-                    parsedip = socket.inet_pton(socket.AF_INET6, ip)
-                    destdata.append(0b10000000)
-                    destdata.extend(parsedip)
-                else:
-                    destdata = None
-                    self.oem_init()
-                    self._oem.set_alert_ipv6_destination(ip, destination,
-                                                         channel)
-            if destdata:
-                self.xraw_command(netfn=0xc, command=1, data=destdata)
+
         if (acknowledge_required is not None
                 or retries is not None
                 or acknowledge_timeout is not None):
@@ -1234,6 +1346,8 @@ class Command(object):
                     currtype[1] |= 0b10000000
                 else:
                     currtype[1] &= 0b1111111
+            # set PET trap destination
+            currtype[1] &= 0b1111000
             if acknowledge_timeout is not None:
                 currtype[2] = acknowledge_timeout
             if retries is not None:
@@ -1241,6 +1355,18 @@ class Command(object):
             destreq = bytearray((channel, 18))
             destreq.extend(currtype)
             self.xraw_command(netfn=0xc, command=1, data=destreq)
+        if ip is not None:
+            destdata = bytearray((channel, 19, destination))
+            try:
+                parsedip = socket.inet_pton(socket.AF_INET, ip)
+                destdata.extend((0, 0))
+                destdata.extend(parsedip)
+                destdata.extend('\x00\x00\x00\x00\x00\x00')
+            except socket.error:
+                parsedip = socket.inet_pton(socket.AF_INET6, ip)
+                destdata.append(0b10000000)
+                destdata.extend(parsedip)
+            self.xraw_command(netfn=0xc, command=1, data=destdata)
         if not ip == '0.0.0.0':
             self._assure_alert_policy(channel, destination)
 
@@ -1262,10 +1388,16 @@ class Command(object):
             return self.get_mci()
 
     def get_mci(self):
-        """Get the Management Controller Identifier, per DCMI specification
+        """Set the management controller identifier.
+
+        Try the OEM command first,if False, then set it per DCMI specification
 
         :returns: The identifier as a string
         """
+        self.oem_init()
+        identifier = self._oem.get_oem_identifier()
+        if identifier:
+            return identifier
         return self._chunkwise_dcmi_fetch(9)
 
     def set_hostname(self, hostname):
@@ -1283,11 +1415,16 @@ class Command(object):
             return self.set_mci(hostname)
 
     def set_mci(self, mci):
-        """Set the management controller identifier, per DCMI specification
+        """Set the management controller identifier.
 
+        Try the OEM command first, if False, then set it per DCMI specification
         """
+        self.oem_init()
         if not isinstance(mci, bytes):
             mci = mci.encode('utf8')
+        ret = self._oem.set_oem_identifier(mci)
+        if ret:
+            return
         return self._chunkwise_dcmi_set(0xa, mci + b'\x00')
 
     def get_asset_tag(self):
@@ -1295,12 +1432,18 @@ class Command(object):
 
         :returns: The asset tag
         """
+        self.oem_init()
+        if hasattr(self._oem, 'get_asset_tag'):
+            return self._oem.get_asset_tag()
         return self._chunkwise_dcmi_fetch(6)
 
     def set_asset_tag(self, tag):
         """Set the asset tag value
 
         """
+        self.oem_init()
+        if hasattr(self._oem, 'set_asset_tag'):
+            return self._oem.set_asset_tag(tag)
         return self._chunkwise_dcmi_set(8, tag)
 
     def _chunkwise_dcmi_fetch(self, command):
@@ -1328,6 +1471,9 @@ class Command(object):
             chunk = bytearray(chunk)
             cmddata = bytearray((0xdc, offset, len(chunk)))
             cmddata += chunk
+            # set offset, otherwise the last setting will override
+            # the previous setting
+            offset += len(chunk)
             self.xraw_command(netfn=0x2c, command=command, data=cmddata)
 
     def set_channel_access(self, channel=None,
@@ -1514,6 +1660,10 @@ class Command(object):
         r['privilege_level'] = privilege_levels[data[1] & 0b00001111]
         return r
 
+    def get_screenshot(self, outfile):
+        self.oem_init()
+        return self._oem.get_screenshot(outfile)
+
     def get_channel_info(self, channel=None):
         """Get channel info
 
@@ -1583,7 +1733,7 @@ class Command(object):
 
         :param channel: number [1:7]
 
-        :parm callback: User Restricted to Callback
+        :param callback: User Restricted to Callback
         False = User Privilege Limit is determined by the User Privilege Limit
             parameter, below, for both callback and non-callback connections.
         True  = User Privilege Limit is determined by the User Privilege Limit
@@ -1621,7 +1771,13 @@ class Command(object):
             * administrator
             * proprietary
             * no_access
+            * custom.<name>
         """
+        self.oem_init()
+        if hasattr(self._oem, 'oem_user_access'):
+            callback, privilege_level = self._oem.oem_user_access(
+                callback, privilege_level)
+
         if channel is None:
             channel = self.get_network_channel()
         b = 0b10000000
@@ -1644,11 +1800,16 @@ class Command(object):
         self.oem_init()
         self._oem.set_user_access(
             uid, channel, callback, link_auth, ipmi_msg, privilege_level)
+        if privilege_level.startswith('custom.'):
+            return True  # unable to proceed with standard support
         data = [b, uid & 0b00111111,
                 privilege_levels[privilege_level] & 0b00001111, 0]
         response = self.raw_command(netfn=0x06, command=0x43, data=data)
         if 'error' in response:
             raise Exception(response['error'])
+        # Set KVM and VMedia Allowed if is administrator
+        if privilege_level == 'administrator':
+            self.set_extended_privilleges(uid)
         return True
 
     def get_user_access(self, uid, channel=None):
@@ -1687,16 +1848,21 @@ class Command(object):
         r['access']['callback'] = (data[3] & 0b01000000) != 0
         r['access']['link_auth'] = (data[3] & 0b00100000) != 0
         r['access']['ipmi_msg'] = (data[3] & 0b00010000) != 0
-        privilege_levels = {
-            0: 'reserved',
-            1: 'callback',
-            2: 'user',
-            3: 'operator',
-            4: 'administrator',
-            5: 'proprietary',
-            0x0F: 'no_access'
-        }
-        r['access']['privilege_level'] = privilege_levels[data[3] & 0b00001111]
+        self.oem_init()
+        oempriv = self._oem.get_user_privilege_level(uid)
+        if oempriv:
+            r['access']['privilege_level'] = oempriv
+        else:
+            privilege_levels = {
+                0: 'reserved',
+                1: 'callback',
+                2: 'user',
+                3: 'operator',
+                4: 'administrator',
+                5: 'proprietary',
+                0x0F: 'no_access'
+            }
+            r['access']['privilege_level'] = privilege_levels[data[3] & 0b00001111]
         return r
 
     def set_user_name(self, uid, name):
@@ -1713,7 +1879,8 @@ class Command(object):
         name = name.ljust(16, b'\x00')
         name = bytearray(name)
         data.extend(name)
-        self.xraw_command(netfn=0x06, command=0x45, data=data)
+        # set timeout to 2s to avoid retry causing enable failure
+        self.xraw_command(netfn=0x06, command=0x45, data=data, timeout=2)
         return True
 
     def get_user_name(self, uid, return_none_on_error=True):
@@ -1776,6 +1943,9 @@ class Command(object):
             else:
                 password = password.ljust(16, b'\x00')
             data.extend(bytearray(password))
+
+        self.oem_init()
+        data = self._oem.process_password(password, data)
         try:
             self.xraw_command(netfn=0x06, command=0x47, data=data)
         except exc.IpmiException as ie:
@@ -1860,13 +2030,14 @@ class Command(object):
                 privilege_level: (str)[callback, user, operatorm administrator,
                                        proprietary, no_access]
         """
+        self.oem_init()
         if channel is None:
             channel = self.get_network_channel()
         names = {}
         max_ids = self.get_channel_max_user_count(channel)
         for uid in range(1, max_ids + 1):
             name = self.get_user_name(uid=uid)
-            if name is not None:
+            if self._oem.is_valid(name):
                 names[uid] = self.get_user(uid=uid, channel=channel)
         return names
 
@@ -1908,6 +2079,17 @@ class Command(object):
         :param channel: number [1:7]
         """
         # TODO(jjohnson2): Provide OEM extensibility to cover user deletion
+        self.oem_init()
+        if hasattr(self._oem, 'user_delete'):
+            try:
+                self._oem.user_delete(uid, channel)
+            except exc.BypassGenericBehavior:
+                return
+        if hasattr(self._oem, 'user_delete_privilege_level'):
+            privilege_level = self._oem.user_delete_privilege_level()
+        else:
+            privilege_level = 'no_access'
+
         if channel is None:
             channel = self.get_network_channel()
         self.set_user_password(uid, mode='disable', password=None)
@@ -1915,18 +2097,14 @@ class Command(object):
         # so new users dont get extra access
         self.set_user_access(uid, channel=channel, callback=False,
                              link_auth=False, ipmi_msg=False,
-                             privilege_level='no_access')
+                             privilege_level=privilege_level)
         try:
             # First try to set name to all \x00 explicitly
+            # Fix bug-174389, don't try to send \xff command,
+            # will cause IMM1 can't login
             self.set_user_name(uid, '')
-        except exc.IpmiException as ie:
-            if ie.ipmicode != 0xcc:
-                raise
-            # An invalid data field in request  is frequently reported.
-            # however another convention that exists is all '\xff'
-            # if this fails, pass up the error so that calling code knows
-            # that the deletion did not go as planned for now
-            self.set_user_name(uid, b'\xff' * 16)
+        except exc.IpmiException:
+            raise
         return True
 
     def disable_user(self, uid, mode):
@@ -1943,6 +2121,51 @@ class Command(object):
         self.set_user_password(uid, mode)
         return True
 
+    def update_user(self, user):
+        """Update User
+
+        Update user attributes, include name, password, access
+
+        :param user: user attributes
+        """
+        if 'username' in user:
+            self.set_user_name(uid=user['uid'], name=user['username'])
+
+        privilege_level = None
+        if 'privilege_level' in user:
+            privilege_level = user['privilege_level']
+
+        if privilege_level and privilege_level == 'no_access':
+            return self.upate_user_no_access(user, privilege_level)
+        else:
+            return self.update_user_default(user, privilege_level)
+
+    def upate_user_no_access(self, user, privilege_level):
+        # if set to no_access, modify password first
+        if 'password' in user:
+            self.set_user_password(uid=user['uid'], password=user['password'])
+        self.set_user_access(uid=user['uid'], privilege_level=privilege_level)
+
+        return True
+
+    def update_user_default(self, user, privilege_level):
+        if privilege_level:
+            self.set_user_access(uid=user['uid'],
+                                 privilege_level=privilege_level)
+        if 'password' in user:
+            self.set_user_password(uid=user['uid'],
+                                   password=user['password'])
+            self.set_user_password(uid=user['uid'], mode='enable',
+                                   password=user['password'])
+        if 'enabled' in user:
+            if user['enabled'] == 'yes':
+                mode = 'enable'
+            else:
+                mode = 'disable'
+            self.disable_user(user['uid'], mode)
+
+        return True
+
     def get_firmware(self, components=()):
         """Retrieve OEM Firmware information"""
 
@@ -1968,6 +2191,14 @@ class Command(object):
         self.oem_init()
         return self._oem.set_oem_capping_enabled(enable)
 
+    def set_server_capping(self, value):
+        self.oem_init()
+        self._oem.set_oem_server_capping(value)
+
+    def get_server_capping(self):
+        self.oem_init()
+        return self._oem.get_oem_server_capping()
+
     def get_remote_kvm_available(self):
         """Get remote KVM availability"""
 
@@ -1993,6 +2224,10 @@ class Command(object):
         self.oem_init()
         return self._oem.get_graphical_console()
 
+    def get_update_status(self):
+        self.oem_init()
+        return self._oem.get_update_status()
+
     def update_firmware(self, filename, data=None, progress=None, bank=None):
         """Send file to BMC to perform firmware update
 
@@ -2028,7 +2263,7 @@ class Command(object):
         self.oem_init()
         return self._oem.detach_remote_media()
 
-    def upload_media(self, filename, progress=None):
+    def upload_media(self, filename, progress=None, data=None):
         """Upload a file to be hosted on the target BMC
 
         This will upload the specified data to
@@ -2040,7 +2275,7 @@ class Command(object):
         :param progress: Optional callback for progress updates
         """
         self.oem_init()
-        return self._oem.upload_media(filename, progress)
+        return self._oem.upload_media(filename, progress, data)
 
     def list_media(self):
         """List attached remote media
@@ -2068,4 +2303,11 @@ class Command(object):
 
     def apply_license(self, filename, progress=None, data=None):
         self.oem_init()
-        return self._oem.apply_license(filename, progress, data=None)
+        return self._oem.apply_license(filename, progress, data)
+
+    def set_extended_privilleges(self, uid):
+        """Set user extended privillege as 'KVM & VMedia Allowed'
+
+        """
+        self.oem_init()
+        return self._oem.set_oem_extended_privilleges(uid)
